#!/usr/bin/env python

"""Console script for qarrayrun.

This script executes a single slot of an array job on an HPC compute node.
It is intended to be used with Sun Grid Engine, SLURM, or Torque job schedulers.
It assumes every instance of the job array runs the same command but with
different arguments.  This script performs the work of looking-up the
arguments in a text file and substituting those arguments into the command
to be executed.

Parameters
----------
The script takes 3 arguments:

1. Name of the environment variable that contains the sub-task number.
   You should use SGE_TASK_ID for grid engine.
   You should use SLURM_ARRAY_TASK_ID for SLURM.
   You should use PBS_ARRAYID for torque.

2. Name of the file containing the arguments for each sub-task with one line
   per sub-task.  This script will extract the arguments for this sub-task
   at the line number identified by the environment variable above.  The
   line is parsed and substituted into the command, replacing the parameter placeholders
   with the actual arguments.

3. The remainder of the command line is the command to be executed with parameter
   placeholders of the form {1}, {2}, {3} ...

Examples
--------
# Sort some txt files, writing the sorted output to new files
ls *.txt > files.txt
echo 'qarrayrun.py SGE_TASK_ID files.txt sort -o sorted.{1} {1}' | qsub -t 1-$(cat files.txt | wc -l) -cwd -j y -V -o log

# Your input file might have multiple columns, use {2} for the 2nd column
# Sort the largest files first
ls *.txt | xargs -n 1 wc -c | sort -n -r > files.txt
echo 'qarrayrun.py SGE_TASK_ID files.txt sort -o sorted.{2} {2}' | qsub -t 1-$(cat files.txt | wc -l) -cwd -j y -V -o log

# Use the --shell option and quote your pipeline when you need shell redirection
# Remove blanks before sorting files
ls *.txt > files.txt
echo 'qarrayrun --shell SGE_TASK_ID files.txt "cat {1} | tr -d [:blank:] | sort > sorted.{1}"' | qsub -t 1-$(cat files.txt | wc -l) -cwd -j y -V -o log
"""

from __future__ import print_function
from __future__ import absolute_import

import argparse
import logging
import sys
import textwrap

from qarrayrun import qarrayrun
from qarrayrun.__init__ import __version__

# Ignore flake8 errors in this module
# flake 8: noqa


def parse_arguments(system_args):
    """Parse command line arguments.

    This is an unusual command line parser -- it only parses the command-line
    arguments it expects, and captures the remaining arguments to pass along
    to the function processing the command.

    Parameters
    ----------
    system_args : list
        List of command line arguments, usually sys.argv[1:].

    Returns
    -------
    Namespace
        Command line arguments are stored as attributes of a Namespace.
    list
        List of remaining argument strings
    """
    description = textwrap.dedent("""
        Executes a single slot of an array job on an HPC computational node. This is
        intended to be used with Sun Grid Engine, SLURM, or Torque job schedulers when
        every instance of the job array runs the same command but with different
        arguments. This script performs the work of looking-up the arguments in a text
        file and substituting those arguments into the command to be executed.""")

    epilog = textwrap.dedent("""
        Examples
        --------
        # Sort some txt files, writing the sorted output to new files
        ls *.txt > files.txt
        echo 'qarrayrun SGE_TASK_ID files.txt sort -o sorted.{1} {1}' | qsub -t 1-$(cat files.txt | wc -l) -cwd -j y -V -o log

        # Your input file might have multiple columns, use {2} for the 2nd column
        # Sort the largest files first
        ls *.txt | xargs -n 1 wc -c | sort -n -r > files.txt
        echo 'qarrayrun SGE_TASK_ID files.txt sort -o sorted.{2} {2}' | qsub -t 1-$(cat files.txt | wc -l) -cwd -j y -V -o log

        # Use the --shell option and quote your pipeline when you need shell redirection
        # Remove blanks before sorting files
        ls *.txt > files.txt
        echo 'qarrayrun --shell SGE_TASK_ID files.txt "cat {1} | tr -d [:blank:] | sort > sorted.{1}"' | qsub -t 1-$(cat files.txt | wc -l) -cwd -j y -V -o log
        """)

    formatter_class = argparse.RawDescriptionHelpFormatter
    parser = argparse.ArgumentParser(description=description, epilog=epilog, formatter_class=formatter_class)

    parser.add_argument(dest="subtask_var", type=str, metavar="NAME", help="""Name of the environment variable that contains the sub-task number.
                                                                              You should use SGE_TASK_ID with Grid Engine, SLURM_ARRAY_TASK_ID with SLURM, and PBS_ARRAYID with Torque.""")
    parser.add_argument(dest="array_file", type=str, help="""Name of the file containing the arguments for each sub-task with one line
                                                             per sub-task.  This script will extract the arguments for this sub-task
                                                             at the line number identified by the sub-task environment variable
                                                             (SGE_TASK_ID, SLURM_ARRAY_TASK_ID, or PBS_ARRAYID).  The line is parsed and substituted
                                                             into the command, replacing the parameter placeholders with the actual
                                                             arguments.""")
    parser.add_argument(dest="command", help="""The remainder of the command line is the command to be executed with parameter
                                                placeholders of the form {1}, {2}, {3} ...""")
    parser.add_argument("--shell", dest="shell", action="store_true", help="Run the command through the shell.")
    parser.add_argument("--version", action="version", version="%(prog)s version " + __version__)

    args, remainder = parser.parse_known_args(system_args)

    return args, remainder


def run_command(args, remainder):
    """Execute a single subtask of an array job.

    Parameters
    ----------
    args : Namespace
        Command line arguments stored as attributes of a Namespace, usually
        parsed from sys.argv
    remainder : list
        List of remaining argument strings
    """
    command = [args.command] + remainder
    command_line = ' '.join(command)
    return qarrayrun.run(args.subtask_var, args.array_file, command_line, args.shell)


def run_command_from_args(args, remainder):
    """Run a command with previously parsed arguments in an argparse namespace.

    This function is intended to be used for unit testing.

    Parameters
    ----------
    args : Namespace
        Command line arguments are stored as attributes of a Namespace.
        The args are obtained by calling parse_argument_list().
    remainder : list
        List of remaining argument strings

    Returns
    -------
    Returns 0 on success if it completes with no exceptions.
    """
    return run_command(args, remainder)


def run_from_line(line):
    """Run a command with a command line.

    This function is intended to be used for unit testing.

    Parameters
    ----------
    line : str
        Command line.

    Returns
    -------
    Returns 0 on success if it completes with no exceptions.
    """
    argv = line.split()
    args, remainder = parse_arguments(argv)
    return run_command(args, remainder)


def main():
    """This is the main function which is turned into an executable
    console script by the setuptools entry_points.  See setup.py.

    To run this function as a script, first install the package:
        $ python setup.py develop
        or
        $ pip install --user qarrayrun

    Parameters
    ----------
    This function must not take any parameters

    Returns
    -------
    The return value is passed to sys.exit().
    """
    enable_log_timestamps = False
    if enable_log_timestamps:
        logging.basicConfig(format="%(asctime)s %(message)s", datefmt="%Y-%m-%d %H:%M:%S", level=logging.INFO)
    else:
        logging.basicConfig(format="%(message)s", level=logging.INFO)
    args, remainder = parse_arguments(sys.argv[1:])
    return run_command(args, remainder)


# This snippet lets you run the cli without installing the entrypoint.
if __name__ == "__main__":
    sys.exit(main())



# -*- coding: utf-8 -*-

"""This module is part of qarrayrun.
"""

from __future__ import print_function
from __future__ import absolute_import

import logging
import os
import re
import shlex
import subprocess


def get_file_line(file_name, line_num):
    """Returns the line at the specified line number from the specified file.

    Parameters
    ----------
    file_name : str
        Path to the file
    line_num : int
        Line number to extract from the file

    Returns
    -------
    line : str
        Line at line_num in file or None if the line_num is larger than the number of lines in the file
    """
    line = ""
    line_counter = 0
    with open(file_name) as f:
        for line in f:
            line_counter += 1
            if line_counter == line_num:
                return line
    return None


def substitute_arguments(command_line, arguments):
    """Replace the parameter placeholders in a command line with actual arguments.

    Parameters
    ----------
    command_line : str
        Command line with parameter placeholders like {1}, {2}, {3}
    arguments : list of str
        List of arguments numbered 1,2,3 ....
        The first argument, arguments[0] corresponds to placeholder {1}.

    Returns
    -------
    command_line : str
        Command line with actual arguments ready for execution

    Examples
    --------
    >>> substitute_arguments("cmd {0}/{1}/{2} -- {3}{4}", ["aa", "bb", "cc"])
    'cmd /aa/bb -- cc'
    """
    args = [""]  # put an empty string at index 0
    args.extend(arguments)

    # Get a list of all the parameter numbers appearing in the command line
    param_nums = re.findall("{([0-9]+)}", command_line)
    param_nums = [int(param_num) for param_num in param_nums]

    # Replace the parameters with actual arguments
    for param_num in param_nums:
        placeholder = "{%s}" % param_num
        if param_num == 0 or param_num >= len(args):
            command_line = command_line.replace(placeholder, "")
        else:
            command_line = command_line.replace(placeholder, args[param_num])

    return command_line


def run(subtask_var, array_file, command_line, shell_flag=False):
    """Execute a single subtask of an array job.

    Lookup the parameters for this subtask, substitute those parameters
    into the command line, and run the command.

    Parameters
    ----------
    subtask_var : str
        Name of the environment variable containing the subtask number.  This
        variable is created automatically by the HPC job execution software
        (Grid Engine, SLURM, or Torque).
    array_file : str
        Path to the file containing array job parameters with one line per array-job
        subtask.  This file must exist on a file system that is accessible to the
        compute nodes.
    command_line : str
        Command line with parameter placeholders like {1}, {2}, {3}
    shell_flag : bool, optional
        Boolean flag indicates if the command should be run in a subshell. Defaults to false.

    Returns
    -------
    int
        Return code of the command after it finishes execution
    """
    # Which sub-task number am I?
    subtask_num = os.environ.get(subtask_var)
    if not subtask_num:
        logging.error("Error: the %s environment variable is not defined." % subtask_var)
        exit(1)

    try:
        subtask_num = int(subtask_num)
    except ValueError:
        logging.error("Error: the %s environment variable does not contain a line number." % subtask_var)
        exit(1)

    if subtask_num <= 0:
        logging.error("Error: the subtask number is %d.  It must be greater than zero." % subtask_num)

    # Verify file exists
    if not os.path.isfile(array_file):
        logging.error("Error: the parameter file %s does not exist." % array_file)
        exit(1)

    # Read and parse the substitution arguments from the input file
    line = get_file_line(array_file, subtask_num)
    if not line:
        exit(0)  # Silently ignore attempts to process beyond the end of the file
    arguments = line.split()

    # Build the command with substituted arguments
    command_line = substitute_arguments(command_line, arguments)

    # Execute the command
    if shell_flag:
        return_code = subprocess.call(command_line, shell=True)
    else:
        command_split = shlex.split(command_line)
        return_code = subprocess.call(command_split, shell=False)
    return return_code



#!/usr/bin/env python
# -*- coding: utf-8 -*-


try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read().replace('.. :changelog:', '')

requirements = [
    # TODO: put package requirements here
]

test_requirements = [
    "pytest",
]

setup(
    name='qarrayrun',
    version='1.1.0',
    description="A helper tool for running array jobs on an HPC computational node.",
    long_description=readme + '\n\n' + history,
    author="Steve Davis",
    author_email='steven.davis@fda.hhs.gov',
    url='https://github.com/CFSAN-Biostatistics/qarrayrun',
    packages=[
        'qarrayrun',
    ],
    package_dir={'qarrayrun':
                 'qarrayrun'},
    include_package_data=True,
    install_requires=requirements,
    license="BSD",
    zip_safe=False,
    keywords=['bioinformatics', 'NGS', 'qarrayrun'],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    entry_points={'console_scripts': ['qarrayrun = qarrayrun.cli:main']},
    setup_requires=["pytest-runner"],
    tests_require=test_requirements
)



# -*- coding: utf-8 -*-

"""
test_qarrayrun
----------------------------------

Tests for `qarrayrun` module.
"""

from __future__ import print_function

import os

from qarrayrun import qarrayrun


def test_run():
    # Create array job parameter file
    from tempfile import NamedTemporaryFile
    f = NamedTemporaryFile(delete=False, mode='w')
    arg_file = f.name
    print("A B C", file=f)
    print("Argument1 Argument2 Argument3", file=f)
    f.close()
    f = NamedTemporaryFile(delete=False, mode='w')
    out_file = f.name

    # Write the arguments in reverse order to out_file
    cmd = "python -c '"
    cmd += 'f = open("%s", "w");' % out_file
    cmd += 'f.write("{3} {2} {1}"); f.close()'
    cmd += "'"
    os.environ["SGE_TASK_ID"] = "2"
    return_code = qarrayrun.run("SGE_TASK_ID", arg_file, cmd)
    assert(return_code == 0)

    # Read the file just created to verify reverse order
    f = open(out_file)
    s = f.read()
    f.close()
    assert(s == "Argument3 Argument2 Argument1")

    # Verify non-zero exit code can be returned
    return_code = qarrayrun.run("SGE_TASK_ID", arg_file, "exit 100", shell_flag=True)
    assert(return_code == 100)

    # Clean up temp files
    os.unlink(arg_file)
    os.unlink(out_file)
